@@ -116,36 +116,13 @@ grd_hwaccel_nvidia_get_cuda_damage_kernels (GrdHwAccelNvidia *hwaccel_nvidia,
   *cu_simplify_dmg_arr = hwaccel_nvidia->cu_simplify_dmg_arr;
 }
 
-static const char *
-get_cuda_error_string (GrdHwAccelNvidia *hwaccel_nvidia,
-                       CUresult          cu_result)
-{
-  CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
-  const char *error_string = NULL;
-  CUresult local_cu_result;
-
-  g_assert (cuda_funcs);
-  g_assert (cuda_funcs->cuGetErrorString);
-
-  local_cu_result = cuda_funcs->cuGetErrorString (cu_result, &error_string);
-  if (G_UNLIKELY (local_cu_result != CUDA_SUCCESS))
-    g_warning ("[HWAccel.CUDA] cuGetErrorString() failed: %i", local_cu_result);
-
-  return error_string;
-}
-
 void
 grd_hwaccel_nvidia_push_cuda_context (GrdHwAccelNvidia *hwaccel_nvidia)
 {
   CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
-  CUresult cu_result;
 
-  cu_result = cuda_funcs->cuCtxPushCurrent (hwaccel_nvidia->cu_context);
-  if (cu_result != CUDA_SUCCESS)
-    {
-      g_error ("[HWAccel.CUDA] Failed to push CUDA context: %s",
-               get_cuda_error_string (hwaccel_nvidia, cu_result));
-    }
+  if (cuda_funcs->cuCtxPushCurrent (hwaccel_nvidia->cu_context) != CUDA_SUCCESS)
+    g_error ("[HWAccel.CUDA] Failed to push CUDA context");
 }
 
 void
@@ -162,15 +139,12 @@ grd_hwaccel_nvidia_register_read_only_gl_buffer (GrdHwAccelNvidia   *hwaccel_nvi
                                                  uint32_t            buffer)
 {
   ExtraCudaFunctions *extra_cuda_funcs = hwaccel_nvidia->extra_cuda_funcs;
-  CUresult cu_result;
 
-  cu_result =
-    extra_cuda_funcs->cuGraphicsGLRegisterBuffer (cuda_resource, buffer,
-                                                  CU_GRAPHICS_REGISTER_FLAGS_READ_ONLY);
-  if (cu_result != CUDA_SUCCESS)
+  if (extra_cuda_funcs->cuGraphicsGLRegisterBuffer (
+        cuda_resource, buffer,
+        CU_GRAPHICS_REGISTER_FLAGS_READ_ONLY) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to register GL buffer: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to register GL buffer");
       return FALSE;
     }
 
@@ -195,22 +169,17 @@ grd_hwaccel_nvidia_map_cuda_resource (GrdHwAccelNvidia   *hwaccel_nvidia,
 {
   CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
   ExtraCudaFunctions *extra_cuda_funcs = hwaccel_nvidia->extra_cuda_funcs;
-  CUresult cu_result;
 
-  cu_result = cuda_funcs->cuGraphicsMapResources (1, &cuda_resource,
-                                                  cuda_stream);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuGraphicsMapResources (1, &cuda_resource,
+                                          cuda_stream) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to map resources: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to map resources");
       return FALSE;
     }
-  cu_result = extra_cuda_funcs->cuGraphicsResourceGetMappedPointer (dev_ptr, size,
-                                                                    cuda_resource);
-  if (cu_result != CUDA_SUCCESS)
+  if (extra_cuda_funcs->cuGraphicsResourceGetMappedPointer (dev_ptr, size,
+                                                            cuda_resource) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to get mapped pointer: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to get mapped pointer");
       cuda_funcs->cuGraphicsUnmapResources (1, &cuda_resource, cuda_stream);
       return FALSE;
     }
@@ -232,13 +201,11 @@ grd_hwaccel_nvidia_create_cuda_stream (GrdHwAccelNvidia *hwaccel_nvidia,
                                        CUstream         *cuda_stream)
 {
   CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
-  CUresult cu_result;
 
-  cu_result = cuda_funcs->cuStreamCreate (cuda_stream, CU_STREAM_NON_BLOCKING);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuStreamCreate (cuda_stream,
+                                  CU_STREAM_NON_BLOCKING) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to create stream: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to create stream");
       return FALSE;
     }
 
@@ -259,13 +226,10 @@ grd_hwaccel_nvidia_alloc_mem (GrdHwAccelNvidia *hwaccel_nvidia,
                               size_t            size)
 {
   CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
-  CUresult cu_result;
 
-  cu_result = cuda_funcs->cuMemAlloc (device_ptr, size);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuMemAlloc (device_ptr, size) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to allocate memory: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to allocate memory");
       return FALSE;
     }
 
@@ -297,18 +261,6 @@ get_next_free_encode_session_id (GrdHwAccelNvidia *hwaccel_nvidia)
   return encode_session_id;
 }
 
-static const char *
-get_last_nvenc_error_string (GrdHwAccelNvidia *hwaccel_nvidia,
-                             void             *encoder)
-{
-  NV_ENCODE_API_FUNCTION_LIST *nvenc_api = &hwaccel_nvidia->nvenc_api;
-
-  g_assert (nvenc_api);
-  g_assert (nvenc_api->nvEncGetLastErrorString);
-
-  return nvenc_api->nvEncGetLastErrorString (encoder);
-}
-
 gboolean
 grd_hwaccel_nvidia_create_nvenc_session (GrdHwAccelNvidia *hwaccel_nvidia,
                                          uint32_t         *encode_session_id,
@@ -340,8 +292,7 @@ grd_hwaccel_nvidia_create_nvenc_session (GrdHwAccelNvidia *hwaccel_nvidia,
   if (hwaccel_nvidia->nvenc_api.nvEncOpenEncodeSessionEx (
         &open_params, &encode_session->encoder) != NV_ENC_SUCCESS)
     {
-      g_debug ("[HWAccel.NVENC] Failed to open encode session: %s",
-               get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_debug ("[HWAccel.NVENC] Failed to open encode session");
       g_free (encode_session);
       return FALSE;
     }
@@ -375,8 +326,7 @@ grd_hwaccel_nvidia_create_nvenc_session (GrdHwAccelNvidia *hwaccel_nvidia,
     {
       NV_ENC_PIC_PARAMS pic_params = {0};
 
-      g_warning ("[HWAccel.NVENC] Failed to initialize encoder: %s",
-                 get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_warning ("[HWAccel.NVENC] Failed to initialize encoder");
       pic_params.encodePicFlags = NV_ENC_PIC_FLAG_EOS;
       hwaccel_nvidia->nvenc_api.nvEncEncodePicture (encode_session->encoder,
                                                     &pic_params);
@@ -392,8 +342,7 @@ grd_hwaccel_nvidia_create_nvenc_session (GrdHwAccelNvidia *hwaccel_nvidia,
     {
       NV_ENC_PIC_PARAMS pic_params = {0};
 
-      g_warning ("[HWAccel.NVENC] Failed to create bitstream buffer: %s",
-                 get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_warning ("[HWAccel.NVENC] Failed to create bitstream buffer");
       pic_params.encodePicFlags = NV_ENC_PIC_FLAG_EOS;
       hwaccel_nvidia->nvenc_api.nvEncEncodePicture (encode_session->encoder,
                                                     &pic_params);
@@ -448,17 +397,16 @@ grd_hwaccel_nvidia_free_nvenc_session (GrdHwAccelNvidia *hwaccel_nvidia,
 }
 
 gboolean
-grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia *hwaccel_nvidia,
-                                             uint32_t          encode_session_id,
-                                             CUdeviceptr       src_data,
-                                             CUdeviceptr      *main_view_nv12,
-                                             uint16_t          src_width,
-                                             uint16_t          src_height,
-                                             uint16_t          aligned_width,
-                                             uint16_t          aligned_height,
-                                             CUstream          cuda_stream)
+grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia  *hwaccel_nvidia,
+                                             uint32_t           encode_session_id,
+                                             CUdeviceptr        src_data,
+                                             CUdeviceptr       *main_view_nv12,
+                                             uint16_t           src_width,
+                                             uint16_t           src_height,
+                                             uint16_t           aligned_width,
+                                             uint16_t           aligned_height,
+                                             CUstream           cuda_stream)
 {
-  CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
   NvEncEncodeSession *encode_session;
   NV_ENC_REGISTER_RESOURCE register_res = {0};
   NV_ENC_MAP_INPUT_RESOURCE map_input_res = {0};
@@ -466,7 +414,6 @@ grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia *hwaccel_nvidia,
   unsigned int grid_dim_x, grid_dim_y, grid_dim_z;
   unsigned int block_dim_x, block_dim_y, block_dim_z;
   void *args[7];
-  CUresult cu_result;
 
   if (!g_hash_table_lookup_extended (hwaccel_nvidia->encode_sessions,
                                      GUINT_TO_POINTER (encode_session_id),
@@ -503,22 +450,17 @@ grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia *hwaccel_nvidia,
   args[5] = &aligned_height;
   args[6] = &aligned_width;
 
-  cu_result = cuda_funcs->cuLaunchKernel (hwaccel_nvidia->cu_bgrx_to_yuv420,
-                                          grid_dim_x, grid_dim_y, grid_dim_z,
-                                          block_dim_x, block_dim_y, block_dim_z,
-                                          0, cuda_stream, args, NULL);
-  if (cu_result != CUDA_SUCCESS)
+  if (hwaccel_nvidia->cuda_funcs->cuLaunchKernel (
+        hwaccel_nvidia->cu_bgrx_to_yuv420, grid_dim_x, grid_dim_y, grid_dim_z,
+        block_dim_x, block_dim_y, block_dim_z, 0, cuda_stream, args, NULL) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to launch BGRX_TO_YUV420 kernel: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to launch BGRX_TO_YUV420 kernel");
       return FALSE;
     }
 
-  cu_result = cuda_funcs->cuStreamSynchronize (cuda_stream);
-  if (cu_result != CUDA_SUCCESS)
+  if (hwaccel_nvidia->cuda_funcs->cuStreamSynchronize (cuda_stream) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to synchronize stream: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to synchronize stream");
       return FALSE;
     }
 
@@ -534,8 +476,7 @@ grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia *hwaccel_nvidia,
   if (hwaccel_nvidia->nvenc_api.nvEncRegisterResource (
         encode_session->encoder, &register_res) != NV_ENC_SUCCESS)
     {
-      g_warning ("[HWAccel.NVENC] Failed to register resource: %s",
-                 get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_warning ("[HWAccel.NVENC] Failed to register resource");
       return FALSE;
     }
 
@@ -545,8 +486,7 @@ grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia *hwaccel_nvidia,
   if (hwaccel_nvidia->nvenc_api.nvEncMapInputResource (
         encode_session->encoder, &map_input_res) != NV_ENC_SUCCESS)
     {
-      g_warning ("[HWAccel.NVENC] Failed to map input resource: %s",
-                 get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_warning ("[HWAccel.NVENC] Failed to map input resource");
       hwaccel_nvidia->nvenc_api.nvEncUnregisterResource (encode_session->encoder,
                                                          register_res.registeredResource);
       return FALSE;
@@ -564,8 +504,7 @@ grd_hwaccel_nvidia_avc420_encode_bgrx_frame (GrdHwAccelNvidia *hwaccel_nvidia,
   if (hwaccel_nvidia->nvenc_api.nvEncEncodePicture (
         encode_session->encoder, &pic_params) != NV_ENC_SUCCESS)
     {
-      g_warning ("[HWAccel.NVENC] Failed to encode frame: %s",
-                 get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_warning ("[HWAccel.NVENC] Failed to encode frame");
       hwaccel_nvidia->nvenc_api.nvEncUnmapInputResource (encode_session->encoder,
                                                          map_input_res.mappedResource);
       hwaccel_nvidia->nvenc_api.nvEncUnregisterResource (encode_session->encoder,
@@ -604,8 +543,7 @@ grd_hwaccel_nvidia_avc420_retrieve_bitstream (GrdHwAccelNvidia  *hwaccel_nvidia,
   if (nvenc_api->nvEncLockBitstream (encode_session->encoder,
                                      &lock_bitstream) != NV_ENC_SUCCESS)
     {
-      g_warning ("[HWAccel.NVENC] Failed to lock bitstream: %s",
-                 get_last_nvenc_error_string (hwaccel_nvidia, encode_session->encoder));
+      g_warning ("[HWAccel.NVENC] Failed to lock bitstream");
       goto out;
     }
 
@@ -759,13 +697,10 @@ load_cuda_module (GrdHwAccelNvidia *hwaccel_nvidia,
                   const char       *ptx_instructions)
 {
   CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
-  CUresult cu_result;
 
-  cu_result = cuda_funcs->cuModuleLoadData (module, ptx_instructions);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuModuleLoadData (module, ptx_instructions) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to load %s module: %s",
-                 name, get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to load %s module", name);
       return FALSE;
     }
 
@@ -779,13 +714,10 @@ load_cuda_function (GrdHwAccelNvidia *hwaccel_nvidia,
                     const char       *name)
 {
   CudaFunctions *cuda_funcs = hwaccel_nvidia->cuda_funcs;
-  CUresult cu_result;
 
-  cu_result = cuda_funcs->cuModuleGetFunction (function, module, name);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuModuleGetFunction (function, module, name) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to get kernel %s: %s",
-                 name, get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to get kernel %s", name);
       return FALSE;
     }
 
@@ -807,7 +739,6 @@ grd_hwaccel_nvidia_new (GrdEglThread *egl_thread)
   g_autofree char *avc_ptx_path = NULL;
   g_autofree char *avc_ptx_instructions = NULL;
   g_autoptr (GError) error = NULL;
-  CUresult cu_result;
   unsigned int i;
 
   hwaccel_nvidia = g_object_new (GRD_TYPE_HWACCEL_NVIDIA, NULL);
@@ -830,11 +761,9 @@ grd_hwaccel_nvidia_new (GrdEglThread *egl_thread)
   cuda_funcs = hwaccel_nvidia->cuda_funcs;
   nvenc_funcs = hwaccel_nvidia->nvenc_funcs;
 
-  cu_result = cuda_funcs->cuInit (0);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuInit (0) != CUDA_SUCCESS)
     {
-      g_debug ("[HWAccel.CUDA] Failed to initialize CUDA: %s",
-               get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_debug ("[HWAccel.CUDA] Failed to initialize CUDA");
       return NULL;
     }
   if (!get_cuda_devices_from_gl_context (hwaccel_nvidia, egl_thread,
@@ -869,12 +798,10 @@ grd_hwaccel_nvidia_new (GrdEglThread *egl_thread)
     }
 
   hwaccel_nvidia->cu_device = cu_device;
-  cu_result = cuda_funcs->cuDevicePrimaryCtxRetain (&hwaccel_nvidia->cu_context,
-                                                    hwaccel_nvidia->cu_device);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuDevicePrimaryCtxRetain (&hwaccel_nvidia->cu_context,
+                                            hwaccel_nvidia->cu_device) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to retain CUDA context: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to retain CUDA context");
       return NULL;
     }
 
@@ -886,11 +813,9 @@ grd_hwaccel_nvidia_new (GrdEglThread *egl_thread)
       return NULL;
     }
 
-  cu_result = cuda_funcs->cuCtxPushCurrent (hwaccel_nvidia->cu_context);
-  if (cu_result != CUDA_SUCCESS)
+  if (cuda_funcs->cuCtxPushCurrent (hwaccel_nvidia->cu_context) != CUDA_SUCCESS)
     {
-      g_warning ("[HWAccel.CUDA] Failed to push CUDA context: %s",
-                 get_cuda_error_string (hwaccel_nvidia, cu_result));
+      g_warning ("[HWAccel.CUDA] Failed to push CUDA context");
       cuda_funcs->cuDevicePrimaryCtxRelease (hwaccel_nvidia->cu_device);
       return NULL;
     }
@@ -957,7 +882,8 @@ grd_hwaccel_nvidia_dispose (GObject *object)
   nvenc_free_functions (&hwaccel_nvidia->nvenc_funcs);
   cuda_free_functions (&hwaccel_nvidia->cuda_funcs);
 
-  g_assert (g_hash_table_size (hwaccel_nvidia->encode_sessions) == 0);
+  g_assert (!hwaccel_nvidia->encode_sessions ||
+            g_hash_table_size (hwaccel_nvidia->encode_sessions) == 0);
   g_clear_pointer (&hwaccel_nvidia->encode_sessions, g_hash_table_destroy);
 
   G_OBJECT_CLASS (grd_hwaccel_nvidia_parent_class)->dispose (object);
