@@ -183,10 +183,8 @@ update_ping_source (GrdRdpNetworkAutodetection *network_autodetection)
   else
     new_ping_interval_type = PING_INTERVAL_LOW;
 
-  if (network_autodetection->ping_interval == new_ping_interval_type)
-    return;
-
-  if (network_autodetection->ping_source)
+  if (network_autodetection->ping_interval != new_ping_interval_type &&
+      network_autodetection->ping_source)
     {
       g_source_destroy (network_autodetection->ping_source);
       g_clear_pointer (&network_autodetection->ping_source, g_source_unref);
@@ -486,7 +484,8 @@ autodetect_rtt_measure_response (rdpContext *rdp_context,
   g_mutex_unlock (&network_autodetection->consumer_mutex);
 
   g_mutex_lock (&network_autodetection->shutdown_mutex);
-  if (!network_autodetection->in_shutdown && has_rtt_consumer_rdpgfx)
+  if (!network_autodetection->in_shutdown && has_rtt_consumer_rdpgfx &&
+      rdp_peer_context->graphics_pipeline)
     {
       grd_rdp_graphics_pipeline_notify_new_round_trip_time (
         rdp_peer_context->graphics_pipeline, avg_round_trip_time_us);
