@@ -27,7 +27,6 @@
 #include <spa/param/props.h>
 #include <spa/param/format-utils.h>
 #include <spa/param/video/format-utils.h>
-#include <spa/utils/result.h>
 #include <sys/mman.h>
 
 #include "grd-context.h"
@@ -58,12 +57,6 @@ typedef void (* GrdRdpFrameReadyCallback) (GrdRdpPipeWireStream *stream,
                                            gboolean              success,
                                            gpointer              user_data);
 
-typedef struct
-{
-  GMutex buffer_mutex;
-  gboolean is_locked;
-} BufferContext;
-
 struct _GrdRdpFrame
 {
   gatomicrefcount refcount;
@@ -135,11 +128,11 @@ struct _GrdRdpPipeWireStream
   struct pw_stream *pipewire_stream;
   struct spa_hook pipewire_stream_listener;
 
-  GHashTable *pipewire_buffers;
-
   uint32_t src_node_id;
 
   struct spa_video_info_raw spa_format;
+
+  gboolean destroyed;
 };
 
 G_DEFINE_TYPE (GrdRdpPipeWireStream, grd_rdp_pipewire_stream,
@@ -149,57 +142,6 @@ static void grd_rdp_frame_unref (GrdRdpFrame *frame);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdRdpFrame, grd_rdp_frame_unref)
 
-static BufferContext *
-buffer_context_new (void)
-{
-  BufferContext *buffer_context;
-
-  buffer_context = g_new0 (BufferContext, 1);
-  g_mutex_init (&buffer_context->buffer_mutex);
-
-  return buffer_context;
-}
-
-static void
-buffer_context_free (BufferContext *buffer_context)
-{
-  g_mutex_clear (&buffer_context->buffer_mutex);
-
-  g_free (buffer_context);
-}
-
-static void
-acquire_pipewire_buffer_lock (GrdRdpPipeWireStream *stream,
-                              struct pw_buffer     *buffer)
-{
-  BufferContext *buffer_context = NULL;
-
-  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
-                                     NULL, (gpointer *) &buffer_context))
-    g_assert_not_reached ();
-
-  g_mutex_lock (&buffer_context->buffer_mutex);
-  g_assert (!buffer_context->is_locked);
-  buffer_context->is_locked = TRUE;
-}
-
-static void
-maybe_release_pipewire_buffer_lock (GrdRdpPipeWireStream *stream,
-                                    struct pw_buffer     *buffer)
-{
-  BufferContext *buffer_context = NULL;
-
-  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
-                                     NULL, (gpointer *) &buffer_context))
-    g_assert_not_reached ();
-
-  if (!buffer_context->is_locked)
-    return;
-
-  buffer_context->is_locked = FALSE;
-  g_mutex_unlock (&buffer_context->buffer_mutex);
-}
-
 static GrdRdpFrame *
 grd_rdp_frame_new (GrdRdpPipeWireStream     *stream,
                    GrdRdpFrameReadyCallback  callback,
@@ -330,71 +272,6 @@ create_render_source (GrdRdpPipeWireStream *stream,
   g_source_attach (stream->render_source, render_context);
 }
 
-static gboolean
-pipewire_loop_source_prepare (GSource *base,
-                              int     *timeout)
-{
-  *timeout = -1;
-  return FALSE;
-}
-
-static gboolean
-pipewire_loop_source_dispatch (GSource     *source,
-                               GSourceFunc  callback,
-                               gpointer     user_data)
-{
-  GrdPipeWireSource *pipewire_source = (GrdPipeWireSource *) source;
-  int result;
-
-  result = pw_loop_iterate (pipewire_source->pipewire_loop, 0);
-  if (result < 0)
-    g_warning ("pipewire_loop_iterate failed: %s", spa_strerror (result));
-
-  return TRUE;
-}
-
-static void
-pipewire_loop_source_finalize (GSource *source)
-{
-  GrdPipeWireSource *pipewire_source = (GrdPipeWireSource *) source;
-
-  pw_loop_leave (pipewire_source->pipewire_loop);
-  pw_loop_destroy (pipewire_source->pipewire_loop);
-}
-
-static GSourceFuncs pipewire_source_funcs =
-{
-  pipewire_loop_source_prepare,
-  NULL,
-  pipewire_loop_source_dispatch,
-  pipewire_loop_source_finalize
-};
-
-static GrdPipeWireSource *
-create_pipewire_source (void)
-{
-  GrdPipeWireSource *pipewire_source;
-
-  pipewire_source =
-    (GrdPipeWireSource *) g_source_new (&pipewire_source_funcs,
-                                        sizeof (GrdPipeWireSource));
-  pipewire_source->pipewire_loop = pw_loop_new (NULL);
-  if (!pipewire_source->pipewire_loop)
-    {
-      g_source_destroy ((GSource *) pipewire_source);
-      return NULL;
-    }
-
-  g_source_add_unix_fd (&pipewire_source->base,
-                        pw_loop_get_fd (pipewire_source->pipewire_loop),
-                        G_IO_IN | G_IO_ERR);
-
-  pw_loop_enter (pipewire_source->pipewire_loop);
-  g_source_attach (&pipewire_source->base, NULL);
-
-  return pipewire_source;
-}
-
 static void
 on_stream_state_changed (void                 *user_data,
                          enum pw_stream_state  old,
@@ -531,33 +408,6 @@ on_stream_param_changed (void                 *user_data,
                            params, G_N_ELEMENTS (params));
 }
 
-static void
-on_stream_add_buffer (void             *user_data,
-                      struct pw_buffer *buffer)
-{
-  GrdRdpPipeWireStream *stream = user_data;
-
-  g_hash_table_insert (stream->pipewire_buffers, buffer, buffer_context_new ());
-}
-
-static void
-on_stream_remove_buffer (void             *user_data,
-                         struct pw_buffer *buffer)
-{
-  GrdRdpPipeWireStream *stream = user_data;
-  BufferContext *buffer_context = NULL;
-
-  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
-                                     NULL, (gpointer *) &buffer_context))
-    g_assert_not_reached ();
-
-  /* Ensure buffer is not locked any more */
-  g_mutex_lock (&buffer_context->buffer_mutex);
-  g_mutex_unlock (&buffer_context->buffer_mutex);
-
-  g_hash_table_remove (stream->pipewire_buffers, buffer);
-}
-
 static void
 process_mouse_pointer_bitmap (GrdRdpPipeWireStream *stream,
                               struct spa_buffer    *buffer,
@@ -717,10 +567,9 @@ on_framebuffer_ready (gboolean success,
 
 static void
 process_frame_data (GrdRdpPipeWireStream *stream,
-                    struct pw_buffer     *pw_buffer,
+                    struct spa_buffer    *buffer,
                     GrdRdpFrame          *frame)
 {
-  struct spa_buffer *buffer = pw_buffer->buffer;
   GrdRdpFrameReadyCallback callback = frame->callback;
   gpointer user_data = frame->callback_user_data;
   uint32_t drm_format;
@@ -815,7 +664,6 @@ process_frame_data (GrdRdpPipeWireStream *stream,
       realize_buffer_data->hwaccel_nvidia = hwaccel_nvidia;
       realize_buffer_data->rdp_buffer = rdp_buffer;
 
-      acquire_pipewire_buffer_lock (stream, pw_buffer);
       grd_egl_thread_upload (egl_thread,
                              pbo,
                              height,
@@ -897,7 +745,6 @@ process_frame_data (GrdRdpPipeWireStream *stream,
           import_buffer_data->rdp_buffer = rdp_buffer;
         }
 
-      acquire_pipewire_buffer_lock (stream, pw_buffer);
       grd_egl_thread_download (egl_thread,
                                rdp_buffer->pbo,
                                height,
@@ -918,83 +765,9 @@ process_frame_data (GrdRdpPipeWireStream *stream,
                                grd_rdp_frame_ref (g_steal_pointer (&frame)),
                                (GDestroyNotify) grd_rdp_frame_unref);
     }
-  else if (buffer->datas[0].type == SPA_DATA_MemPtr)
-    {
-      GrdSession *session = GRD_SESSION (stream->session_rdp);
-      GrdContext *context = grd_session_get_context (session);
-      GrdEglThread *egl_thread = grd_context_get_egl_thread (context);
-      GrdHwAccelNvidia *hwaccel_nvidia = stream->rdp_surface->hwaccel_nvidia;
-      AllocateBufferData *allocate_buffer_data;
-      RealizeBufferData *realize_buffer_data;
-      GrdRdpBuffer *rdp_buffer;
-      void *src_data;
-      uint32_t pbo;
-      uint8_t *data_to_upload;
-
-      src_data = buffer->datas[0].data;
-
-      frame->buffer = grd_rdp_buffer_pool_acquire (stream->buffer_pool);
-      if (!frame->buffer)
-        {
-          grd_session_rdp_notify_error (stream->session_rdp,
-                                        GRD_SESSION_RDP_ERROR_GRAPHICS_SUBSYSTEM_FAILED);
-          callback (stream, g_steal_pointer (&frame), FALSE, user_data);
-          return;
-        }
-      rdp_buffer = frame->buffer;
-      pbo = rdp_buffer->pbo;
-
-      if (stream->rdp_surface->needs_no_local_data &&
-          src_stride == dst_stride)
-        {
-          data_to_upload = src_data;
-        }
-      else
-        {
-          copy_frame_data (frame,
-                           src_data,
-                           width, height,
-                           dst_stride,
-                           src_stride,
-                           bpp);
-
-          data_to_upload = rdp_buffer->local_data;
-        }
-
-      if (!hwaccel_nvidia)
-        {
-          callback (stream, g_steal_pointer (&frame), TRUE, user_data);
-          return;
-        }
-
-      unmap_cuda_resources (egl_thread, hwaccel_nvidia, rdp_buffer);
-
-      allocate_buffer_data = g_new0 (AllocateBufferData, 1);
-      allocate_buffer_data->hwaccel_nvidia = hwaccel_nvidia;
-      allocate_buffer_data->rdp_buffer = rdp_buffer;
-
-      realize_buffer_data = g_new0 (RealizeBufferData, 1);
-      realize_buffer_data->hwaccel_nvidia = hwaccel_nvidia;
-      realize_buffer_data->rdp_buffer = rdp_buffer;
-
-      grd_egl_thread_upload (egl_thread,
-                             pbo,
-                             height,
-                             dst_stride,
-                             data_to_upload,
-                             cuda_allocate_buffer,
-                             allocate_buffer_data,
-                             g_free,
-                             cuda_map_resource,
-                             realize_buffer_data,
-                             g_free,
-                             on_framebuffer_ready,
-                             grd_rdp_frame_ref (g_steal_pointer (&frame)),
-                             (GDestroyNotify) grd_rdp_frame_unref);
-    }
   else
     {
-      g_assert_not_reached ();
+      callback (stream, g_steal_pointer (&frame), TRUE, user_data);
     }
 }
 
@@ -1053,9 +826,6 @@ out:
 
   g_source_set_ready_time (stream->render_source, 0);
 
-  if (buffer)
-    maybe_release_pipewire_buffer_lock (stream, buffer);
-
   g_clear_pointer (&frame, grd_rdp_frame_unref);
 }
 
@@ -1068,8 +838,23 @@ on_stream_process (void *user_data)
   struct pw_buffer *last_frame_buffer = NULL;
   struct pw_buffer *next_buffer;
 
+  if (stream->destroyed)
+    return;
+
   while ((next_buffer = pw_stream_dequeue_buffer (stream->pipewire_stream)))
     {
+      struct spa_meta_header *spa_meta_header;
+
+      spa_meta_header = spa_buffer_find_meta_data (next_buffer->buffer,
+                                                   SPA_META_Header,
+                                                   sizeof (struct spa_meta_header));
+      if (spa_meta_header &&
+          spa_meta_header->flags & SPA_META_HEADER_FLAG_CORRUPTED)
+        {
+          pw_stream_queue_buffer (stream->pipewire_stream, next_buffer);
+          continue;
+        }
+
       if (grd_pipewire_buffer_has_pointer_bitmap (next_buffer))
         {
           if (last_pointer_buffer == last_frame_buffer)
@@ -1113,7 +898,7 @@ on_stream_process (void *user_data)
       return;
     }
 
-  process_frame_data (stream, last_frame_buffer,
+  process_frame_data (stream, last_frame_buffer->buffer,
                       g_steal_pointer (&frame));
 }
 
@@ -1121,8 +906,6 @@ static const struct pw_stream_events stream_events = {
   PW_VERSION_STREAM_EVENTS,
   .state_changed = on_stream_state_changed,
   .param_changed = on_stream_param_changed,
-  .add_buffer = on_stream_add_buffer,
-  .remove_buffer = on_stream_remove_buffer,
   .process = on_stream_process,
 };
 
@@ -1317,22 +1100,19 @@ grd_rdp_pipewire_stream_new (GrdSessionRdp               *session_rdp,
   g_autoptr (GrdRdpPipeWireStream) stream = NULL;
   GrdPipeWireSource *pipewire_source;
 
-  grd_maybe_initialize_pipewire ();
-
   stream = g_object_new (GRD_TYPE_RDP_PIPEWIRE_STREAM, NULL);
   stream->session_rdp = session_rdp;
   stream->rdp_surface = rdp_surface;
   stream->src_node_id = src_node_id;
 
+  pw_init (NULL, NULL);
+
   create_render_source (stream, render_context);
 
-  pipewire_source = create_pipewire_source ();
+  pipewire_source = grd_attached_pipewire_source_new ("RDP", error);
   if (!pipewire_source)
-    {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                   "Failed to create PipeWire source");
-      return NULL;
-    }
+    return NULL;
+
   stream->pipewire_source = (GSource *) pipewire_source;
 
   stream->pipewire_context = pw_context_new (pipewire_source->pipewire_loop,
@@ -1378,15 +1158,21 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
   GrdContext *context = grd_session_get_context (session);
   GrdEglThread *egl_thread;
 
-  /* Setting a PipeWire stream inactive will wait for the data thread to end */
+  stream->destroyed = TRUE;
+
   if (stream->pipewire_stream)
-    pw_stream_set_active (stream->pipewire_stream, false);
+    pw_stream_flush (stream->pipewire_stream, false);
 
   egl_thread = grd_context_get_egl_thread (context);
   if (egl_thread && stream->pipewire_stream)
     sync_egl_thread (egl_thread);
 
-  g_clear_pointer (&stream->pipewire_stream, pw_stream_destroy);
+  /*
+   * We can't clear stream->pipewire_stream before destroying it, as the data
+   * thread in PipeWire might access the variable during destruction.
+   */
+  if (stream->pipewire_stream)
+    pw_stream_destroy (stream->pipewire_stream);
 
   g_clear_pointer (&stream->pipewire_core, pw_core_disconnect);
   g_clear_pointer (&stream->pipewire_context, pw_context_destroy);
@@ -1410,7 +1196,7 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
 
   g_mutex_clear (&stream->frame_mutex);
 
-  g_clear_pointer (&stream->pipewire_buffers, g_hash_table_unref);
+  pw_deinit ();
 
   G_OBJECT_CLASS (grd_rdp_pipewire_stream_parent_class)->finalize (object);
 }
@@ -1418,10 +1204,6 @@ grd_rdp_pipewire_stream_finalize (GObject *object)
 static void
 grd_rdp_pipewire_stream_init (GrdRdpPipeWireStream *stream)
 {
-  stream->pipewire_buffers =
-    g_hash_table_new_full (NULL, NULL,
-                           NULL, (GDestroyNotify) buffer_context_free);
-
   g_mutex_init (&stream->frame_mutex);
 }
 
