@@ -32,9 +32,11 @@
 #include "grd-context.h"
 #include "grd-damage-utils.h"
 #include "grd-hwaccel-nvidia.h"
+#include "grd-rdp-audio-playback.h"
 #include "grd-rdp-buffer.h"
 #include "grd-rdp-damage-detector.h"
 #include "grd-rdp-display-control.h"
+#include "grd-rdp-dvc.h"
 #include "grd-rdp-event-queue.h"
 #include "grd-rdp-graphics-pipeline.h"
 #include "grd-rdp-network-autodetection.h"
@@ -43,6 +45,7 @@
 #include "grd-rdp-sam.h"
 #include "grd-rdp-server.h"
 #include "grd-rdp-surface.h"
+#include "grd-rdp-telemetry.h"
 #include "grd-settings.h"
 #include "grd-stream.h"
 
@@ -74,6 +77,19 @@ typedef enum _PauseKeyState
   PAUSE_KEY_STATE_CTRL_UP,
 } PauseKeyState;
 
+typedef struct _SessionMetrics
+{
+  int64_t rd_session_start_init_us;
+  int64_t rd_session_ready_us;
+  int64_t rd_session_started_us;
+
+  gboolean received_first_frame;
+  gboolean sent_first_frame;
+  int64_t first_frame_ready_us;
+  int64_t first_frame_sent_us;
+  uint32_t skipped_frames;
+} SessionMetrics;
+
 typedef struct _Pointer
 {
   uint8_t *bitmap;
@@ -123,7 +139,9 @@ struct _GrdSessionRdp
   GrdRdpSAMFile *sam_file;
   uint32_t rdp_error_info;
   GrdRdpScreenShareMode screen_share_mode;
-  gboolean session_stopped;
+  gboolean session_should_stop;
+
+  SessionMetrics session_metrics;
 
   GMutex rdp_flags_mutex;
   RdpPeerFlag rdp_flags;
@@ -269,9 +287,19 @@ take_or_encode_frame (GrdSessionRdp *session_rdp,
                       GrdRdpSurface *rdp_surface,
                       GrdRdpBuffer  *buffer)
 {
+  SessionMetrics *session_metrics = &session_rdp->session_metrics;
   uint16_t width = buffer->width;
   uint16_t height = buffer->height;
 
+  if (!session_metrics->received_first_frame)
+    {
+      session_metrics->first_frame_ready_us = g_get_monotonic_time ();
+      session_metrics->received_first_frame = TRUE;
+      g_debug ("[RDP] Received first frame");
+    }
+  if (!session_metrics->sent_first_frame && rdp_surface->pending_framebuffer)
+    ++session_metrics->skipped_frames;
+
   g_clear_pointer (&rdp_surface->pending_framebuffer, grd_rdp_buffer_release);
   maybe_resize_graphics_output_buffer (session_rdp, width, height);
 
@@ -626,6 +654,7 @@ maybe_queue_close_session_idle (GrdSessionRdp *session_rdp)
     g_idle_add (close_session_idle, session_rdp);
   g_mutex_unlock (&session_rdp->close_session_mutex);
 
+  session_rdp->session_should_stop = TRUE;
   SetEvent (session_rdp->stop_event);
 }
 
@@ -650,6 +679,32 @@ grd_session_rdp_notify_error (GrdSessionRdp      *session_rdp,
   maybe_queue_close_session_idle (session_rdp);
 }
 
+void
+grd_session_rdp_tear_down_channel (GrdSessionRdp *session_rdp,
+                                   GrdRdpChannel  channel)
+{
+  freerdp_peer *peer = session_rdp->peer;
+  RdpPeerContext *rdp_peer_context = (RdpPeerContext *) peer->context;
+
+  g_mutex_lock (&rdp_peer_context->channel_mutex);
+  switch (channel)
+    {
+    case GRD_RDP_CHANNEL_NONE:
+      g_assert_not_reached ();
+      break;
+    case GRD_RDP_CHANNEL_AUDIO_PLAYBACK:
+      g_clear_object (&rdp_peer_context->audio_playback);
+      break;
+    case GRD_RDP_CHANNEL_DISPLAY_CONTROL:
+      g_clear_object (&rdp_peer_context->display_control);
+      break;
+    case GRD_RDP_CHANNEL_TELEMETRY:
+      g_clear_object (&rdp_peer_context->telemetry);
+      break;
+    }
+  g_mutex_unlock (&rdp_peer_context->channel_mutex);
+}
+
 static void
 handle_client_gone (GrdSessionRdp *session_rdp)
 {
@@ -684,7 +739,7 @@ get_current_monitor_layout (GrdSessionRdp  *session_rdp,
   (*monitors)[0].flags = MONITOR_PRIMARY;
 }
 
-static void
+static gboolean
 rdp_peer_refresh_gfx (GrdSessionRdp  *session_rdp,
                       GrdRdpSurface  *rdp_surface,
                       GrdRdpBuffer   *buffer)
@@ -709,11 +764,11 @@ rdp_peer_refresh_gfx (GrdSessionRdp  *session_rdp,
       unset_rdp_peer_flag (session_rdp, RDP_PEER_PENDING_GFX_GRAPHICS_RESET);
     }
 
-  grd_rdp_graphics_pipeline_refresh_gfx (graphics_pipeline,
-                                         rdp_surface, buffer);
+  return grd_rdp_graphics_pipeline_refresh_gfx (graphics_pipeline,
+                                                rdp_surface, buffer);
 }
 
-static void
+static gboolean
 rdp_peer_refresh_rfx (GrdSessionRdp  *session_rdp,
                       GrdRdpSurface  *rdp_surface,
                       cairo_region_t *region,
@@ -808,6 +863,8 @@ rdp_peer_refresh_rfx (GrdSessionRdp  *session_rdp,
 
   g_free (rfx_messages);
   g_free (rfx_rects);
+
+  return TRUE;
 }
 
 static void
@@ -861,7 +918,7 @@ rdp_peer_encode_nsc_rect (gpointer data,
   g_mutex_unlock (thread_pool_context->pending_jobs_mutex);
 }
 
-static void
+static gboolean
 rdp_peer_refresh_nsc (GrdSessionRdp  *session_rdp,
                       GrdRdpSurface  *rdp_surface,
                       cairo_region_t *region,
@@ -956,6 +1013,8 @@ rdp_peer_refresh_nsc (GrdSessionRdp  *session_rdp,
     }
 
   g_free (encode_contexts);
+
+  return TRUE;
 }
 
 static void
@@ -1104,7 +1163,7 @@ rdp_peer_refresh_raw_rect (freerdp_peer          *peer,
     }
 }
 
-static void
+static gboolean
 rdp_peer_refresh_raw (GrdSessionRdp  *session_rdp,
                       GrdRdpSurface  *rdp_surface,
                       cairo_region_t *region,
@@ -1230,6 +1289,27 @@ rdp_peer_refresh_raw (GrdSessionRdp  *session_rdp,
     g_free (bitmap_data[i].bitmapDataStream);
 
   g_free (bitmap_data);
+
+  return TRUE;
+}
+
+static void
+print_session_metrics (SessionMetrics *session_metrics)
+{
+  g_debug ("[RDP] Remote Desktop session metrics: "
+           "_ready: %lims, _started: %lims, firstFrameReceived: %lims, "
+           "firstFrameSent: %lims (skipped_frames: %u), wholeStartup: %lims",
+           (session_metrics->rd_session_ready_us -
+            session_metrics->rd_session_start_init_us) / 1000,
+           (session_metrics->rd_session_started_us -
+            session_metrics->rd_session_ready_us) / 1000,
+           (session_metrics->first_frame_ready_us -
+           session_metrics->rd_session_ready_us) / 1000,
+           (session_metrics->first_frame_sent_us -
+           session_metrics->first_frame_ready_us) / 1000,
+           session_metrics->skipped_frames,
+           (session_metrics->first_frame_sent_us -
+           session_metrics->rd_session_start_init_us) / 1000);
 }
 
 static void
@@ -1237,10 +1317,12 @@ rdp_peer_refresh_region (GrdSessionRdp  *session_rdp,
                          GrdRdpSurface  *rdp_surface,
                          GrdRdpBuffer   *buffer)
 {
+  SessionMetrics *session_metrics = &session_rdp->session_metrics;
   freerdp_peer *peer = session_rdp->peer;
   RdpPeerContext *rdp_peer_context = (RdpPeerContext *) peer->context;
   rdpSettings *rdp_settings = peer->settings;
   cairo_region_t *region = NULL;
+  gboolean success;
 
   if (!rdp_settings->SupportGraphicsPipeline)
     {
@@ -1254,17 +1336,25 @@ rdp_peer_refresh_region (GrdSessionRdp  *session_rdp,
     }
 
   if (rdp_settings->SupportGraphicsPipeline)
-    rdp_peer_refresh_gfx (session_rdp, rdp_surface, buffer);
+    success = rdp_peer_refresh_gfx (session_rdp, rdp_surface, buffer);
   else if (rdp_settings->RemoteFxCodec)
-    rdp_peer_refresh_rfx (session_rdp, rdp_surface, region, buffer);
+    success = rdp_peer_refresh_rfx (session_rdp, rdp_surface, region, buffer);
   else if (rdp_settings->NSCodec)
-    rdp_peer_refresh_nsc (session_rdp, rdp_surface, region, buffer);
+    success = rdp_peer_refresh_nsc (session_rdp, rdp_surface, region, buffer);
   else
-    rdp_peer_refresh_raw (session_rdp, rdp_surface, region, buffer);
+    success = rdp_peer_refresh_raw (session_rdp, rdp_surface, region, buffer);
 
   g_clear_pointer (&region, cairo_region_destroy);
 
   ++rdp_peer_context->frame_id;
+
+  if (success && !session_metrics->sent_first_frame)
+    {
+      session_metrics->first_frame_sent_us = g_get_monotonic_time ();
+      session_metrics->sent_first_frame = TRUE;
+
+      print_session_metrics (session_metrics);
+    }
 }
 
 static gboolean
@@ -1659,9 +1749,8 @@ rdp_peer_capabilities (freerdp_peer *peer)
       session_rdp->monitor_config->monitor_count = 1;
     }
 
-  if ((rdp_settings->SupportGraphicsPipeline || rdp_settings->RemoteFxCodec ||
-       rdp_settings->NSCodec) &&
-      rdp_settings->ColorDepth != 32)
+  if (rdp_settings->SupportGraphicsPipeline || rdp_settings->RemoteFxCodec ||
+      rdp_settings->NSCodec)
     {
       g_debug ("[RDP] Fixing invalid colour depth set by client");
       rdp_settings->ColorDepth = 32;
@@ -1706,9 +1795,12 @@ rdp_peer_post_connect (freerdp_peer *peer)
 {
   RdpPeerContext *rdp_peer_context = (RdpPeerContext *) peer->context;
   GrdSessionRdp *session_rdp = rdp_peer_context->session_rdp;
+  SessionMetrics *session_metrics = &session_rdp->session_metrics;
   rdpSettings *rdp_settings = peer->settings;
 
-  g_debug ("New RDP client");
+  g_debug ("New RDP client: [OS major type, OS minor type]: [%s, %s]",
+           freerdp_peer_os_major_type_string (peer),
+           freerdp_peer_os_minor_type_string (peer));
 
   if (!rdp_settings->SupportGraphicsPipeline &&
       !rdp_settings->RemoteFxCodec &&
@@ -1746,6 +1838,18 @@ rdp_peer_post_connect (freerdp_peer *peer)
                  "(RTT detection, Bandwidth measurement). "
                  "High latency connections will suffer!");
     }
+  if (rdp_settings->AudioPlayback && !rdp_settings->NetworkAutoDetect)
+    {
+      g_warning ("[RDP] Client does not support autodetecting network "
+                 "characteristics. Disabling audio output redirection");
+      rdp_settings->AudioPlayback = FALSE;
+    }
+  if (rdp_settings->AudioPlayback && !rdp_settings->SupportGraphicsPipeline)
+    {
+      g_warning ("[RDP] Client does not support graphics pipeline. Disabling "
+                 "audio output redirection");
+      rdp_settings->AudioPlayback = FALSE;
+    }
 
   if (rdp_settings->NetworkAutoDetect)
     {
@@ -1754,22 +1858,9 @@ rdp_peer_post_connect (freerdp_peer *peer)
     }
 
   if (rdp_settings->SupportGraphicsPipeline)
-    {
-      set_rdp_peer_flag (session_rdp, RDP_PEER_PENDING_GFX_INIT);
-
-      rdp_peer_context->graphics_pipeline =
-        grd_rdp_graphics_pipeline_new (session_rdp,
-                                       session_rdp->graphics_context,
-                                       rdp_peer_context->vcm,
-                                       session_rdp->stop_event,
-                                       peer->context,
-                                       rdp_peer_context->network_autodetection,
-                                       rdp_peer_context->encode_stream,
-                                       rdp_peer_context->rfx_context);
-      grd_rdp_graphics_pipeline_set_hwaccel_nvidia (
-        rdp_peer_context->graphics_pipeline, session_rdp->hwaccel_nvidia);
-    }
+    set_rdp_peer_flag (session_rdp, RDP_PEER_PENDING_GFX_INIT);
 
+  session_metrics->rd_session_start_init_us = g_get_monotonic_time ();
   grd_session_start (GRD_SESSION (session_rdp));
 
   g_clear_pointer (&session_rdp->sam_file, grd_rdp_sam_free_sam_file);
@@ -1797,7 +1888,8 @@ rdp_peer_activate (freerdp_peer *peer)
 
   g_debug ("Activating client");
 
-  if (rdp_peer_context->graphics_pipeline &&
+  if ((rdp_peer_context->graphics_pipeline ||
+       is_rdp_peer_flag_set (session_rdp, RDP_PEER_PENDING_GFX_INIT)) &&
       !rdp_settings->SupportGraphicsPipeline)
     {
       g_warning ("Client disabled support for the graphics pipeline during the "
@@ -1826,8 +1918,9 @@ rdp_peer_context_free (freerdp_peer   *peer,
   if (!rdp_peer_context)
     return;
 
-  if (rdp_peer_context->vcm != INVALID_HANDLE_VALUE)
-    g_clear_pointer (&rdp_peer_context->vcm, WTSCloseServer);
+  g_clear_object (&rdp_peer_context->rdp_dvc);
+
+  g_clear_pointer (&rdp_peer_context->vcm, WTSCloseServer);
 
   if (rdp_peer_context->encode_stream)
     {
@@ -1836,6 +1929,8 @@ rdp_peer_context_free (freerdp_peer   *peer,
     }
 
   g_clear_pointer (&rdp_peer_context->rfx_context, rfx_context_free);
+
+  g_mutex_clear (&rdp_peer_context->channel_mutex);
 }
 
 static BOOL
@@ -1846,10 +1941,13 @@ rdp_peer_context_new (freerdp_peer   *peer,
 
   rdp_peer_context->frame_id = 0;
 
+  g_mutex_init (&rdp_peer_context->channel_mutex);
+
   rdp_peer_context->rfx_context = rfx_context_new (TRUE);
   if (!rdp_peer_context->rfx_context)
     {
       g_warning ("[RDP] Failed to create RFX context");
+      rdp_peer_context_free (peer, rdp_peer_context);
       return FALSE;
     }
   rfx_context_set_pixel_format (rdp_peer_context->rfx_context,
@@ -1869,13 +1967,16 @@ rdp_peer_context_new (freerdp_peer   *peer,
   rdp_peer_context->planar_flags |= PLANAR_FORMAT_HEADER_RLE;
 
   rdp_peer_context->vcm = WTSOpenServerA ((LPSTR) peer->context);
-  if (!rdp_peer_context->vcm || rdp_peer_context->vcm == INVALID_HANDLE_VALUE)
+  if (!rdp_peer_context->vcm)
     {
-      g_warning ("[RDP] Failed to create virtual channel manager");
+      g_warning ("[RDP] Failed to retrieve VCM handle");
       rdp_peer_context_free (peer, rdp_peer_context);
       return FALSE;
     }
 
+  rdp_peer_context->rdp_dvc = grd_rdp_dvc_new (rdp_peer_context->vcm,
+                                               &rdp_peer_context->rdp_context);
+
   return TRUE;
 }
 
@@ -1920,6 +2021,7 @@ init_rdp_session (GrdSessionRdp  *session_rdp,
       freerdp_peer_free (peer);
       return FALSE;
     }
+  session_rdp->peer = peer;
 
   rdp_peer_context = (RdpPeerContext *) peer->context;
   rdp_peer_context->session_rdp = session_rdp;
@@ -1929,8 +2031,6 @@ init_rdp_session (GrdSessionRdp  *session_rdp,
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
                    "Failed to create SAM database");
-      freerdp_peer_context_free (peer);
-      freerdp_peer_free (peer);
       return FALSE;
     }
 
@@ -1940,19 +2040,20 @@ init_rdp_session (GrdSessionRdp  *session_rdp,
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
                    "Failed to set path of SAM database");
-      freerdp_peer_context_free (peer);
-      freerdp_peer_free (peer);
       return FALSE;
     }
 
-  rdp_settings->CertificateFile = strdup (grd_settings_get_rdp_server_cert (settings));
-  rdp_settings->PrivateKeyFile = strdup (grd_settings_get_rdp_server_key (settings));
+  rdp_settings->CertificateFile =
+    g_strdup (grd_settings_get_rdp_server_cert (settings));
+  rdp_settings->PrivateKeyFile =
+    g_strdup (grd_settings_get_rdp_server_key (settings));
   rdp_settings->RdpSecurity = FALSE;
   rdp_settings->TlsSecurity = FALSE;
   rdp_settings->NlaSecurity = TRUE;
 
   rdp_settings->OsMajorType = OSMAJORTYPE_UNIX;
   rdp_settings->OsMinorType = OSMINORTYPE_PSEUDO_XSERVER;
+  rdp_settings->AudioPlayback = TRUE;
   rdp_settings->ColorDepth = 32;
   rdp_settings->GfxAVC444v2 = rdp_settings->GfxAVC444 = FALSE;
   rdp_settings->GfxH264 = FALSE;
@@ -1962,6 +2063,7 @@ init_rdp_session (GrdSessionRdp  *session_rdp,
   rdp_settings->HasHorizontalWheel = TRUE;
   rdp_settings->NetworkAutoDetect = TRUE;
   rdp_settings->RefreshRect = TRUE;
+  rdp_settings->RemoteConsoleAudio = TRUE;
   rdp_settings->RemoteFxCodec = TRUE;
   rdp_settings->SupportGraphicsPipeline = TRUE;
   rdp_settings->NSCodec = TRUE;
@@ -1986,14 +2088,9 @@ init_rdp_session (GrdSessionRdp  *session_rdp,
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
                    "Failed to initialize peer");
-      g_clear_pointer (&session_rdp->sam_file, grd_rdp_sam_free_sam_file);
-      freerdp_peer_context_free (peer);
-      freerdp_peer_free (peer);
       return FALSE;
     }
 
-  session_rdp->peer = peer;
-
   return TRUE;
 }
 
@@ -2003,7 +2100,6 @@ socket_thread_func (gpointer data)
   GrdSessionRdp *session_rdp = data;
   freerdp_peer *peer;
   RdpPeerContext *rdp_peer_context;
-  rdpSettings *rdp_settings;
   HANDLE vcm;
   HANDLE channel_event;
   HANDLE events[32];
@@ -2015,7 +2111,6 @@ socket_thread_func (gpointer data)
 
   peer = session_rdp->peer;
   rdp_peer_context = (RdpPeerContext *) peer->context;
-  rdp_settings = peer->settings;
   vcm = rdp_peer_context->vcm;
   channel_event = WTSVirtualChannelManagerGetEventHandle (vcm);
 
@@ -2036,6 +2131,8 @@ socket_thread_func (gpointer data)
           events[n_events++] = bw_measure_stop_event;
         }
 
+      events[n_events++] = channel_event;
+
       n_freerdp_handles = peer->GetEventHandles (peer, &events[n_events],
                                                  32 - n_events);
       if (!n_freerdp_handles)
@@ -2046,11 +2143,9 @@ socket_thread_func (gpointer data)
         }
       n_events += n_freerdp_handles;
 
-      events[n_events++] = channel_event;
-
       WaitForMultipleObjects (n_events, events, FALSE, INFINITE);
 
-      if (WaitForSingleObject (session_rdp->stop_event, 0) == WAIT_OBJECT_0)
+      if (session_rdp->session_should_stop)
         break;
 
       if (!peer->CheckFileDescriptor (peer))
@@ -2062,7 +2157,9 @@ socket_thread_func (gpointer data)
 
       if (WTSVirtualChannelManagerIsChannelJoined (vcm, "drdynvc"))
         {
+          GrdRdpTelemetry *telemetry;
           GrdRdpGraphicsPipeline *graphics_pipeline;
+          GrdRdpAudioPlayback *audio_playback;
           GrdRdpDisplayControl *display_control;
 
           switch (WTSVirtualChannelManagerGetDrdynvcState (vcm))
@@ -2075,17 +2172,25 @@ socket_thread_func (gpointer data)
               SetEvent (channel_event);
               break;
             case DRDYNVC_STATE_READY:
+              g_mutex_lock (&rdp_peer_context->channel_mutex);
+              telemetry = rdp_peer_context->telemetry;
               graphics_pipeline = rdp_peer_context->graphics_pipeline;
+              audio_playback = rdp_peer_context->audio_playback;
               display_control = rdp_peer_context->display_control;
 
-              if (rdp_settings->SupportGraphicsPipeline)
+              if (telemetry && !session_rdp->session_should_stop)
+                grd_rdp_telemetry_maybe_init (telemetry);
+              if (graphics_pipeline && !session_rdp->session_should_stop)
                 grd_rdp_graphics_pipeline_maybe_init (graphics_pipeline);
-
-              grd_rdp_display_control_maybe_init (display_control);
+              if (audio_playback && !session_rdp->session_should_stop)
+                grd_rdp_audio_playback_maybe_init (audio_playback);
+              if (display_control && !session_rdp->session_should_stop)
+                grd_rdp_display_control_maybe_init (display_control);
+              g_mutex_unlock (&rdp_peer_context->channel_mutex);
               break;
             }
 
-          if (WaitForSingleObject (session_rdp->stop_event, 0) == WAIT_OBJECT_0)
+          if (session_rdp->session_should_stop)
             break;
         }
 
@@ -2121,7 +2226,7 @@ graphics_thread_func (gpointer data)
   if (session_rdp->hwaccel_nvidia)
     grd_hwaccel_nvidia_push_cuda_context (session_rdp->hwaccel_nvidia);
 
-  while (!session_rdp->session_stopped)
+  while (!session_rdp->session_should_stop)
     g_main_context_iteration (session_rdp->graphics_context, TRUE);
 
   if (session_rdp->hwaccel_nvidia)
@@ -2144,27 +2249,17 @@ grd_session_rdp_new (GrdRdpServer      *rdp_server,
 
   context = grd_rdp_server_get_context (rdp_server);
   settings = grd_context_get_settings (context);
-  username = grd_settings_get_rdp_username (settings, &error);
-  if (!username)
+  if (!grd_settings_get_rdp_credentials (settings,
+                                         &username, &password,
+                                         &error))
     {
       if (error)
-        g_warning ("[RDP] Couldn't retrieve RDP username: %s", error->message);
+        g_warning ("[RDP] Couldn't retrieve RDP credentials: %s", error->message);
       else
-        g_message ("[RDP] Username is not set, denying client");
+        g_message ("[RDP] Credentials are not set, denying client");
 
       return NULL;
     }
-  password = grd_settings_get_rdp_password (settings, &error);
-  if (!password)
-    {
-      if (error)
-        g_warning ("[RDP] Couldn't retrieve RDP password: %s", error->message);
-      else
-        g_message ("[RDP] Password is not set, denying client");
-
-      g_free (username);
-      return NULL;
-    }
 
   session_rdp = g_object_new (GRD_TYPE_SESSION_RDP,
                               "context", context,
@@ -2173,7 +2268,7 @@ grd_session_rdp_new (GrdRdpServer      *rdp_server,
   session_rdp->connection = g_object_ref (connection);
   session_rdp->hwaccel_nvidia = hwaccel_nvidia;
 
-  session_rdp->screen_share_mode = grd_settings_get_screen_share_mode (settings);
+  session_rdp->screen_share_mode = grd_settings_get_rdp_screen_share_mode (settings);
 
   if (!init_rdp_session (session_rdp, username, password, &error))
     {
@@ -2197,6 +2292,18 @@ grd_session_rdp_new (GrdRdpServer      *rdp_server,
   return g_steal_pointer (&session_rdp);
 }
 
+static gboolean
+has_session_close_queued (GrdSessionRdp *session_rdp)
+{
+  gboolean pending_session_closing;
+
+  g_mutex_lock (&session_rdp->close_session_mutex);
+  pending_session_closing = session_rdp->close_session_idle_id != 0;
+  g_mutex_unlock (&session_rdp->close_session_mutex);
+
+  return pending_session_closing;
+}
+
 static void
 clear_session_sources (GrdSessionRdp *session_rdp)
 {
@@ -2212,6 +2319,18 @@ clear_session_sources (GrdSessionRdp *session_rdp)
     }
 }
 
+static void
+clear_rdp_peer (GrdSessionRdp *session_rdp)
+{
+  g_clear_pointer (&session_rdp->sam_file, grd_rdp_sam_free_sam_file);
+
+  if (session_rdp->peer)
+    {
+      freerdp_peer_context_free (session_rdp->peer);
+      g_clear_pointer (&session_rdp->peer, freerdp_peer_free);
+    }
+}
+
 static gboolean
 clear_pointer_bitmap (gpointer key,
                       gpointer value,
@@ -2235,14 +2354,13 @@ grd_session_rdp_stop (GrdSession *session)
   g_debug ("Stopping RDP session");
 
   unset_rdp_peer_flag (session_rdp, RDP_PEER_ACTIVATED);
-  session_rdp->session_stopped = TRUE;
+  session_rdp->session_should_stop = TRUE;
+  SetEvent (session_rdp->stop_event);
 
-  if (WaitForSingleObject (session_rdp->stop_event, 0) == WAIT_TIMEOUT)
+  if (!has_session_close_queued (session_rdp))
     {
       freerdp_set_error_info (peer->context->rdp,
                               ERRINFO_RPC_INITIATED_DISCONNECT);
-
-      SetEvent (session_rdp->stop_event);
     }
   else if (session_rdp->rdp_error_info)
     {
@@ -2258,7 +2376,7 @@ grd_session_rdp_stop (GrdSession *session)
   if (session_rdp->graphics_thread)
     {
       g_assert (session_rdp->graphics_context);
-      g_assert (session_rdp->session_stopped);
+      g_assert (session_rdp->session_should_stop);
 
       g_main_context_wakeup (session_rdp->graphics_context);
       g_clear_pointer (&session_rdp->graphics_thread, g_thread_join);
@@ -2267,8 +2385,10 @@ grd_session_rdp_stop (GrdSession *session)
   g_clear_object (&session_rdp->pipewire_stream);
 
   g_clear_object (&rdp_peer_context->clipboard_rdp);
+  g_clear_object (&rdp_peer_context->audio_playback);
   g_clear_object (&rdp_peer_context->display_control);
   g_clear_object (&rdp_peer_context->graphics_pipeline);
+  g_clear_object (&rdp_peer_context->telemetry);
 
   g_clear_pointer (&session_rdp->socket_thread, g_thread_join);
   clear_session_sources (session_rdp);
@@ -2276,16 +2396,13 @@ grd_session_rdp_stop (GrdSession *session)
   peer->Close (peer);
   g_clear_object (&session_rdp->connection);
 
-  g_clear_pointer (&session_rdp->sam_file, grd_rdp_sam_free_sam_file);
-
   g_clear_object (&rdp_peer_context->network_autodetection);
 
   if (session_rdp->thread_pool)
     g_thread_pool_free (session_rdp->thread_pool, FALSE, TRUE);
 
   peer->Disconnect (peer);
-  freerdp_peer_context_free (peer);
-  freerdp_peer_free (peer);
+  clear_rdp_peer (session_rdp);
 
   g_hash_table_foreach_remove (session_rdp->pressed_keys,
                                notify_keycode_released,
@@ -2317,27 +2434,86 @@ close_session_idle (gpointer user_data)
   return G_SOURCE_REMOVE;
 }
 
+static void
+maybe_initialize_graphics_pipeline (GrdSessionRdp *session_rdp)
+{
+  freerdp_peer *peer = session_rdp->peer;
+  RdpPeerContext *rdp_peer_context = (RdpPeerContext *) peer->context;
+  rdpSettings *rdp_settings = peer->settings;
+  GrdRdpGraphicsPipeline *graphics_pipeline;
+  GrdRdpTelemetry *telemetry;
+
+  g_assert (!rdp_peer_context->telemetry);
+  g_assert (!rdp_peer_context->graphics_pipeline);
+
+  if (!rdp_settings->SupportGraphicsPipeline)
+    return;
+
+  g_assert (is_rdp_peer_flag_set (session_rdp, RDP_PEER_PENDING_GFX_INIT));
+
+  telemetry = grd_rdp_telemetry_new (session_rdp,
+                                     rdp_peer_context->rdp_dvc,
+                                     rdp_peer_context->vcm,
+                                     peer->context);
+  rdp_peer_context->telemetry = telemetry;
+
+  graphics_pipeline =
+    grd_rdp_graphics_pipeline_new (session_rdp,
+                                   rdp_peer_context->rdp_dvc,
+                                   session_rdp->graphics_context,
+                                   rdp_peer_context->vcm,
+                                   peer->context,
+                                   rdp_peer_context->network_autodetection,
+                                   rdp_peer_context->encode_stream,
+                                   rdp_peer_context->rfx_context);
+  grd_rdp_graphics_pipeline_set_hwaccel_nvidia (graphics_pipeline,
+                                                session_rdp->hwaccel_nvidia);
+  rdp_peer_context->graphics_pipeline = graphics_pipeline;
+}
+
+static void
+initialize_remaining_virtual_channels (GrdSessionRdp *session_rdp)
+{
+  freerdp_peer *peer = session_rdp->peer;
+  RdpPeerContext *rdp_peer_context = (RdpPeerContext *) peer->context;
+  rdpSettings *rdp_settings = peer->settings;
+  GrdRdpDvc *rdp_dvc = rdp_peer_context->rdp_dvc;
+  HANDLE vcm = rdp_peer_context->vcm;
+
+  if (WTSVirtualChannelManagerIsChannelJoined (vcm, "cliprdr"))
+    {
+      gboolean peer_is_on_ms_windows;
+
+      peer_is_on_ms_windows = rdp_settings->OsMajorType == OSMAJORTYPE_WINDOWS;
+      rdp_peer_context->clipboard_rdp =
+        grd_clipboard_rdp_new (session_rdp, vcm, !peer_is_on_ms_windows);
+    }
+  if (rdp_settings->AudioPlayback && !rdp_settings->RemoteConsoleAudio)
+    {
+      rdp_peer_context->audio_playback =
+        grd_rdp_audio_playback_new (session_rdp, rdp_dvc, vcm, peer->context);
+    }
+}
+
 static void
 grd_session_rdp_remote_desktop_session_ready (GrdSession *session)
 {
   GrdSessionRdp *session_rdp = GRD_SESSION_RDP (session);
-  freerdp_peer *peer = session_rdp->peer;
-  RdpPeerContext *rdp_peer_context = (RdpPeerContext *) peer->context;
+  SessionMetrics *session_metrics = &session_rdp->session_metrics;
 
-  if (WTSVirtualChannelManagerIsChannelJoined (rdp_peer_context->vcm,
-                                               "cliprdr"))
-    {
-      rdp_peer_context->clipboard_rdp =
-        grd_clipboard_rdp_new (session_rdp,
-                               rdp_peer_context->vcm,
-                               session_rdp->stop_event);
-    }
+  session_metrics->rd_session_ready_us = g_get_monotonic_time ();
+
+  maybe_initialize_graphics_pipeline (session_rdp);
+  initialize_remaining_virtual_channels (session_rdp);
 }
 
 static void
 grd_session_rdp_remote_desktop_session_started (GrdSession *session)
 {
   GrdSessionRdp *session_rdp = GRD_SESSION_RDP (session);
+  SessionMetrics *session_metrics = &session_rdp->session_metrics;
+
+  session_metrics->rd_session_started_us = g_get_monotonic_time ();
 
   if (session_rdp->monitor_config->is_virtual)
     g_debug ("[RDP] Remote Desktop session will use virtual monitors");
@@ -2412,8 +2588,8 @@ grd_session_rdp_stream_ready (GrdSession *session,
         {
           rdp_peer_context->display_control =
             grd_rdp_display_control_new (session_rdp,
+                                         rdp_peer_context->rdp_dvc,
                                          rdp_peer_context->vcm,
-                                         session_rdp->stop_event,
                                          MAX_MONITOR_COUNT);
         }
     }
@@ -2445,6 +2621,7 @@ grd_session_rdp_dispose (GObject *object)
   GrdSessionRdp *session_rdp = GRD_SESSION_RDP (object);
 
   clear_session_sources (session_rdp);
+  clear_rdp_peer (session_rdp);
 
   g_assert (!session_rdp->graphics_thread);
   g_clear_pointer (&session_rdp->graphics_context, g_main_context_unref);
