@@ -23,16 +23,15 @@
 #include "grd-vnc-pipewire-stream.h"
 
 #include <drm_fourcc.h>
-#include <pipewire-0.3/pipewire/stream.h>
 #include <pipewire/pipewire.h>
 #include <spa/param/props.h>
 #include <spa/param/format-utils.h>
-#include <spa/utils/result.h>
 #include <sys/mman.h>
 
 #include "grd-context.h"
 #include "grd-egl-thread.h"
 #include "grd-pipewire-utils.h"
+#include "grd-utils.h"
 #include "grd-vnc-cursor.h"
 
 enum
@@ -51,12 +50,6 @@ typedef void (* GrdVncFrameReadyCallback) (GrdVncPipeWireStream *stream,
                                            gboolean              success,
                                            gpointer              user_data);
 
-typedef struct
-{
-  GMutex buffer_mutex;
-  gboolean is_locked;
-} BufferContext;
-
 struct _GrdVncFrame
 {
   gatomicrefcount refcount;
@@ -91,11 +84,11 @@ struct _GrdVncPipeWireStream
   struct pw_stream *pipewire_stream;
   struct spa_hook pipewire_stream_listener;
 
-  GHashTable *pipewire_buffers;
-
   uint32_t src_node_id;
 
   struct spa_video_info_raw spa_format;
+
+  gboolean destroyed;
 };
 
 G_DEFINE_TYPE (GrdVncPipeWireStream, grd_vnc_pipewire_stream,
@@ -105,120 +98,28 @@ static void grd_vnc_frame_unref (GrdVncFrame *frame);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC (GrdVncFrame, grd_vnc_frame_unref)
 
-static gboolean
-pipewire_loop_source_prepare (GSource *base,
-                              int     *timeout)
+void
+grd_vnc_pipewire_stream_resize (GrdVncPipeWireStream *stream,
+                                GrdVncVirtualMonitor *virtual_monitor)
 {
-  *timeout = -1;
-  return FALSE;
-}
+  struct spa_rectangle virtual_monitor_rect;
+  uint8_t params_buffer[1024];
+  struct spa_pod_builder pod_builder;
+  const struct spa_pod *params[1];
 
-static gboolean
-pipewire_loop_source_dispatch (GSource     *source,
-                               GSourceFunc  callback,
-                               gpointer     user_data)
-{
-  GrdPipeWireSource *pipewire_source = (GrdPipeWireSource *) source;
-  int result;
+  virtual_monitor_rect = SPA_RECTANGLE (virtual_monitor->width,
+                                        virtual_monitor->height);
 
-  result = pw_loop_iterate (pipewire_source->pipewire_loop, 0);
-  if (result < 0)
-    g_warning ("pipewire_loop_iterate failed: %s", spa_strerror (result));
+  pod_builder = SPA_POD_BUILDER_INIT (params_buffer, sizeof (params_buffer));
 
-  return TRUE;
-}
+  params[0] = spa_pod_builder_add_object (
+    &pod_builder,
+    SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
+    SPA_FORMAT_VIDEO_size, SPA_POD_Rectangle (&virtual_monitor_rect),
+    0);
 
-static BufferContext *
-buffer_context_new (void)
-{
-  BufferContext *buffer_context;
-
-  buffer_context = g_new0 (BufferContext, 1);
-  g_mutex_init (&buffer_context->buffer_mutex);
-
-  return buffer_context;
-}
-
-static void
-buffer_context_free (BufferContext *buffer_context)
-{
-  g_mutex_clear (&buffer_context->buffer_mutex);
-
-  g_free (buffer_context);
-}
-
-static void
-acquire_pipewire_buffer_lock (GrdVncPipeWireStream *stream,
-                              struct pw_buffer     *buffer)
-{
-  BufferContext *buffer_context = NULL;
-
-  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
-                                     NULL, (gpointer *) &buffer_context))
-    g_assert_not_reached ();
-
-  g_mutex_lock (&buffer_context->buffer_mutex);
-  g_assert (!buffer_context->is_locked);
-  buffer_context->is_locked = TRUE;
-}
-
-static void
-maybe_release_pipewire_buffer_lock (GrdVncPipeWireStream *stream,
-                                    struct pw_buffer     *buffer)
-{
-  BufferContext *buffer_context = NULL;
-
-  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
-                                     NULL, (gpointer *) &buffer_context))
-    g_assert_not_reached ();
-
-  if (!buffer_context->is_locked)
-    return;
-
-  buffer_context->is_locked = FALSE;
-  g_mutex_unlock (&buffer_context->buffer_mutex);
-}
-
-static void
-pipewire_loop_source_finalize (GSource *source)
-{
-  GrdPipeWireSource *pipewire_source = (GrdPipeWireSource *) source;
-
-  pw_loop_leave (pipewire_source->pipewire_loop);
-  pw_loop_destroy (pipewire_source->pipewire_loop);
-}
-
-static GSourceFuncs pipewire_source_funcs =
-{
-  pipewire_loop_source_prepare,
-  NULL,
-  pipewire_loop_source_dispatch,
-  pipewire_loop_source_finalize
-};
-
-static GrdPipeWireSource *
-create_pipewire_source (void)
-{
-  GrdPipeWireSource *pipewire_source;
-
-  pipewire_source =
-    (GrdPipeWireSource *) g_source_new (&pipewire_source_funcs,
-                                        sizeof (GrdPipeWireSource));
-  pipewire_source->pipewire_loop = pw_loop_new (NULL);
-  if (!pipewire_source->pipewire_loop)
-    {
-      g_source_destroy ((GSource *) pipewire_source);
-      return NULL;
-    }
-
-  g_source_add_unix_fd (&pipewire_source->base,
-                        pw_loop_get_fd (pipewire_source->pipewire_loop),
-                        G_IO_IN | G_IO_ERR);
-
-  pw_loop_enter (pipewire_source->pipewire_loop);
-  g_source_attach (&pipewire_source->base, NULL);
-
-  return pipewire_source;
+  pw_stream_update_params (stream->pipewire_stream,
+                           params, G_N_ELEMENTS (params));
 }
 
 static void
@@ -272,6 +173,9 @@ on_stream_param_changed (void                 *user_data,
   width = stream->spa_format.size.width;
   height = stream->spa_format.size.height;
 
+  g_debug ("[VNC] Stream parameters changed. New monitor size: [%u, %u]",
+           width, height);
+
   grd_session_vnc_queue_resize_framebuffer (stream->session, width, height);
 
   allowed_buffer_types = 1 << SPA_DATA_MemFd;
@@ -305,33 +209,6 @@ on_stream_param_changed (void                 *user_data,
                            params, G_N_ELEMENTS (params));
 }
 
-static void
-on_stream_add_buffer (void             *user_data,
-                      struct pw_buffer *buffer)
-{
-  GrdVncPipeWireStream *stream = user_data;
-
-  g_hash_table_insert (stream->pipewire_buffers, buffer, buffer_context_new ());
-}
-
-static void
-on_stream_remove_buffer (void             *user_data,
-                         struct pw_buffer *buffer)
-{
-  GrdVncPipeWireStream *stream = user_data;
-  BufferContext *buffer_context = NULL;
-
-  if (!g_hash_table_lookup_extended (stream->pipewire_buffers, buffer,
-                                     NULL, (gpointer *) &buffer_context))
-    g_assert_not_reached ();
-
-  /* Ensure buffer is not locked any more */
-  g_mutex_lock (&buffer_context->buffer_mutex);
-  g_mutex_unlock (&buffer_context->buffer_mutex);
-
-  g_hash_table_remove (stream->pipewire_buffers, buffer);
-}
-
 static GrdVncFrame *
 grd_vnc_frame_new (GrdVncPipeWireStream     *stream,
                    GrdVncFrameReadyCallback  callback,
@@ -494,10 +371,9 @@ on_dma_buf_downloaded (gboolean success,
 
 static void
 process_frame_data (GrdVncPipeWireStream *stream,
-                    struct pw_buffer     *pw_buffer,
+                    struct spa_buffer    *buffer,
                     GrdVncFrame          *frame)
 {
-  struct spa_buffer *buffer = pw_buffer->buffer;
   GrdVncFrameReadyCallback callback = frame->callback;
   gpointer user_data = frame->callback_user_data;
   int dst_stride;
@@ -527,7 +403,7 @@ process_frame_data (GrdVncPipeWireStream *stream,
       if (map == MAP_FAILED)
         {
           g_warning ("Failed to mmap buffer: %s", g_strerror (errno));
-          callback (stream, g_steal_pointer (&frame), FALSE, user_data);
+          callback (stream, NULL, FALSE, user_data);
           return;
         }
 
@@ -574,9 +450,8 @@ process_frame_data (GrdVncPipeWireStream *stream,
             modifiers[i] = stream->spa_format.modifier;
         }
       dst_data = g_malloc0 (height * dst_stride);
-      frame->data = dst_data;
 
-      acquire_pipewire_buffer_lock (stream, pw_buffer);
+      frame->data = dst_data;
       grd_egl_thread_download (egl_thread,
                                0, 0, 0,
                                NULL, NULL, NULL,
@@ -593,24 +468,9 @@ process_frame_data (GrdVncPipeWireStream *stream,
                                grd_vnc_frame_ref (g_steal_pointer (&frame)),
                                (GDestroyNotify) grd_vnc_frame_unref);
     }
-  else if (buffer->datas[0].type == SPA_DATA_MemPtr)
-    {
-      uint8_t *src_data;
-      int src_stride;
-
-      src_data = buffer->datas[0].data;
-
-      src_stride = buffer->datas[0].chunk->stride;
-      copy_frame_data (frame, src_data,
-                       width, height,
-                       dst_stride, src_stride,
-                       bpp);
-
-      callback (stream, g_steal_pointer (&frame), TRUE, user_data);
-    }
   else
     {
-      g_assert_not_reached ();
+      callback (stream, NULL, FALSE, user_data);
     }
 }
 
@@ -672,9 +532,6 @@ out:
 
   g_source_set_ready_time (stream->pending_frame_source, 0);
 
-  if (buffer)
-    maybe_release_pipewire_buffer_lock (stream, buffer);
-
   g_clear_pointer (&frame, grd_vnc_frame_unref);
 }
 
@@ -708,8 +565,23 @@ on_stream_process (void *user_data)
   int cursor_x = 0;
   int cursor_y = 0;
 
+  if (stream->destroyed)
+    return;
+
   while ((next_buffer = pw_stream_dequeue_buffer (stream->pipewire_stream)))
     {
+      struct spa_meta_header *spa_meta_header;
+
+      spa_meta_header = spa_buffer_find_meta_data (next_buffer->buffer,
+                                                   SPA_META_Header,
+                                                   sizeof (struct spa_meta_header));
+      if (spa_meta_header &&
+          spa_meta_header->flags & SPA_META_HEADER_FLAG_CORRUPTED)
+        {
+          pw_stream_queue_buffer (stream->pipewire_stream, next_buffer);
+          continue;
+        }
+
       maybe_consume_pointer_position (next_buffer, &cursor_moved,
                                       &cursor_x, &cursor_y);
 
@@ -760,7 +632,7 @@ on_stream_process (void *user_data)
       return;
     }
 
-  process_frame_data (stream, last_frame_buffer,
+  process_frame_data (stream, last_frame_buffer->buffer,
                       g_steal_pointer (&frame));
 }
 
@@ -768,14 +640,13 @@ static const struct pw_stream_events stream_events = {
   PW_VERSION_STREAM_EVENTS,
   .state_changed = on_stream_state_changed,
   .param_changed = on_stream_param_changed,
-  .add_buffer = on_stream_add_buffer,
-  .remove_buffer = on_stream_remove_buffer,
   .process = on_stream_process,
 };
 
 static void
-add_common_format_params (struct spa_pod_builder *pod_builder,
-                          enum spa_video_format   spa_format)
+add_common_format_params (struct spa_pod_builder     *pod_builder,
+                          enum spa_video_format       spa_format,
+                          const GrdVncVirtualMonitor *virtual_monitor)
 {
   struct spa_rectangle min_rect;
   struct spa_rectangle max_rect;
@@ -796,11 +667,26 @@ add_common_format_params (struct spa_pod_builder *pod_builder,
   spa_pod_builder_add (pod_builder,
                        SPA_FORMAT_VIDEO_format,
                        SPA_POD_Id (spa_format), 0);
-  spa_pod_builder_add (pod_builder,
-                       SPA_FORMAT_VIDEO_size,
-                       SPA_POD_CHOICE_RANGE_Rectangle (&min_rect,
-                                                       &min_rect,
-                                                       &max_rect), 0);
+
+  if (virtual_monitor)
+    {
+      struct spa_rectangle virtual_monitor_rect;
+
+      virtual_monitor_rect = SPA_RECTANGLE (virtual_monitor->width,
+                                            virtual_monitor->height);
+      spa_pod_builder_add (pod_builder,
+                           SPA_FORMAT_VIDEO_size,
+                           SPA_POD_Rectangle (&virtual_monitor_rect), 0);
+    }
+  else
+    {
+      spa_pod_builder_add (pod_builder,
+                           SPA_FORMAT_VIDEO_size,
+                           SPA_POD_CHOICE_RANGE_Rectangle (&min_rect,
+                                                           &min_rect,
+                                                           &max_rect), 0);
+    }
+
   spa_pod_builder_add (pod_builder,
                        SPA_FORMAT_VIDEO_framerate,
                        SPA_POD_Fraction (&SPA_FRACTION (0, 1)), 0);
@@ -812,8 +698,9 @@ add_common_format_params (struct spa_pod_builder *pod_builder,
 }
 
 static gboolean
-connect_to_stream (GrdVncPipeWireStream  *stream,
-                   GError               **error)
+connect_to_stream (GrdVncPipeWireStream        *stream,
+                   const GrdVncVirtualMonitor  *virtual_monitor,
+                   GError                     **error)
 {
   GrdSession *session = GRD_SESSION (stream->session);
   GrdContext *context = grd_session_get_context (session);
@@ -836,7 +723,7 @@ connect_to_stream (GrdVncPipeWireStream  *stream,
   spa_pod_builder_push_object (&pod_builder, &format_frame,
                                SPA_TYPE_OBJECT_Format,
                                SPA_PARAM_EnumFormat);
-  add_common_format_params (&pod_builder, spa_format);
+  add_common_format_params (&pod_builder, spa_format, virtual_monitor);
 
   egl_thread = grd_context_get_egl_thread (context);
   if (egl_thread)
@@ -882,7 +769,7 @@ connect_to_stream (GrdVncPipeWireStream  *stream,
       spa_pod_builder_push_object (&pod_builder, &format_frame,
                                    SPA_TYPE_OBJECT_Format,
                                    SPA_PARAM_EnumFormat);
-      add_common_format_params (&pod_builder, spa_format);
+      add_common_format_params (&pod_builder, spa_format, virtual_monitor);
       params[1] = spa_pod_builder_pop (&pod_builder, &format_frame);
     }
 
@@ -930,27 +817,25 @@ static const struct pw_core_events core_events = {
 };
 
 GrdVncPipeWireStream *
-grd_vnc_pipewire_stream_new (GrdSessionVnc  *session_vnc,
-                             uint32_t        src_node_id,
-                             GError        **error)
+grd_vnc_pipewire_stream_new (GrdSessionVnc               *session_vnc,
+                             uint32_t                     src_node_id,
+                             const GrdVncVirtualMonitor  *virtual_monitor,
+                             GError                     **error)
 {
   g_autoptr (GrdVncPipeWireStream) stream = NULL;
   GrdPipeWireSource *pipewire_source;
   GSource *source;
 
-  grd_maybe_initialize_pipewire ();
-
   stream = g_object_new (GRD_TYPE_VNC_PIPEWIRE_STREAM, NULL);
   stream->session = session_vnc;
   stream->src_node_id = src_node_id;
 
-  pipewire_source = create_pipewire_source ();
+  pw_init (NULL, NULL);
+
+  pipewire_source = grd_attached_pipewire_source_new ("VNC", error);
   if (!pipewire_source)
-    {
-      g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                   "Failed to create PipeWire source");
-      return NULL;
-    }
+    return NULL;
+
   stream->pipewire_source = (GSource *) pipewire_source;
 
   stream->pipewire_context = pw_context_new (pipewire_source->pipewire_loop,
@@ -958,7 +843,7 @@ grd_vnc_pipewire_stream_new (GrdSessionVnc  *session_vnc,
   if (!stream->pipewire_context)
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                   "Failed to create pipewire context");
+                   "Failed to create PipeWire context");
       return NULL;
     }
 
@@ -966,7 +851,7 @@ grd_vnc_pipewire_stream_new (GrdSessionVnc  *session_vnc,
   if (!stream->pipewire_core)
     {
       g_set_error (error, G_IO_ERROR, G_IO_ERROR_FAILED,
-                   "Failed to connect pipewire context");
+                   "Failed to connect PipeWire context");
       return NULL;
     }
 
@@ -981,29 +866,19 @@ grd_vnc_pipewire_stream_new (GrdSessionVnc  *session_vnc,
                         &core_events,
                         stream);
 
-  if (!connect_to_stream (stream, error))
+  if (!connect_to_stream (stream, virtual_monitor, error))
     return NULL;
 
   return g_steal_pointer (&stream);
 }
 
-typedef struct
-{
-  GMutex mutex;
-  GCond cond;
-  gboolean done;
-} SyncData;
-
 static void
-on_sync_done (gboolean success,
-              gpointer user_data)
+on_sync_complete (gboolean success,
+                  gpointer user_data)
 {
-  SyncData *sync_data = user_data;
+  GrdSyncPoint *sync_point = user_data;
 
-  g_mutex_lock (&sync_data->mutex);
-  sync_data->done = TRUE;
-  g_cond_signal (&sync_data->cond);
-  g_mutex_unlock (&sync_data->mutex);
+  grd_sync_point_complete (sync_point, success);
 }
 
 static void
@@ -1014,30 +889,29 @@ grd_vnc_pipewire_stream_finalize (GObject *object)
   GrdContext *context = grd_session_get_context (session);
   GrdEglThread *egl_thread;
 
-  /* Setting a PipeWire stream inactive will wait for the data thread to end */
+  stream->destroyed = TRUE;
+
   if (stream->pipewire_stream)
-    pw_stream_set_active (stream->pipewire_stream, false);
+    pw_stream_flush (stream->pipewire_stream, false);
 
   egl_thread = grd_context_get_egl_thread (context);
   if (egl_thread)
     {
-      SyncData sync_data = {};
+      GrdSyncPoint sync_point = {};
 
-      g_mutex_init (&sync_data.mutex);
-      g_cond_init (&sync_data.cond);
+      grd_sync_point_init (&sync_point);
+      grd_egl_thread_sync (egl_thread, on_sync_complete, &sync_point, NULL);
 
-      grd_egl_thread_sync (egl_thread, on_sync_done, &sync_data, NULL);
-
-      g_mutex_lock (&sync_data.mutex);
-      while (!sync_data.done)
-        g_cond_wait (&sync_data.cond, &sync_data.mutex);
-      g_mutex_unlock (&sync_data.mutex);
-
-      g_cond_clear (&sync_data.cond);
-      g_mutex_clear (&sync_data.mutex);
+      grd_sync_point_wait_for_completion (&sync_point);
+      grd_sync_point_clear (&sync_point);
     }
 
-  g_clear_pointer (&stream->pipewire_stream, pw_stream_destroy);
+  /*
+   * We can't clear stream->pipewire_stream before destroying it, as the data
+   * thread in PipeWire might access the variable during destruction.
+   */
+  if (stream->pipewire_stream)
+    pw_stream_destroy (stream->pipewire_stream);
 
   g_clear_pointer (&stream->pipewire_core, pw_core_disconnect);
   g_clear_pointer (&stream->pipewire_context, pw_context_destroy);
@@ -1052,7 +926,7 @@ grd_vnc_pipewire_stream_finalize (GObject *object)
 
   g_mutex_clear (&stream->frame_mutex);
 
-  g_clear_pointer (&stream->pipewire_buffers, g_hash_table_unref);
+  pw_deinit ();
 
   G_OBJECT_CLASS (grd_vnc_pipewire_stream_parent_class)->finalize (object);
 }
@@ -1060,10 +934,6 @@ grd_vnc_pipewire_stream_finalize (GObject *object)
 static void
 grd_vnc_pipewire_stream_init (GrdVncPipeWireStream *stream)
 {
-  stream->pipewire_buffers =
-    g_hash_table_new_full (NULL, NULL,
-                           NULL, (GDestroyNotify) buffer_context_free);
-
   g_mutex_init (&stream->frame_mutex);
 }
 
